from pathlib import Path
from typing import Any, Optional, Tuple
from fastapi.responses import JSONResponse
import numpy as np
import pandas as pd
import json
from uiwiz.element import Element
from enum import Enum
import html

JS_PATH = Path(__file__).parent / "aggrid.js"


class Aggrid(Element):
    class OPTIONS(str, Enum):
        autoSizeColumn = "autoSizeAll"
        fitColumnContent = "sizeToFit"

    _classes: str = "ag-theme-alpine"

    def __init__(self, df: pd.DataFrame, column_option: OPTIONS = OPTIONS.autoSizeColumn) -> None:
        super().__init__("div", libraries=["/static/libs/aggrid-community.min.js"], extension=JS_PATH)
        self.classes(Aggrid._classes)

        cols, rows = Aggrid.create_cols_and_rows(df)

        self.attributes["style"] = "width: 100%;"

        self.attributes["hx-ext"] = "hx-aggrid"
        self.attributes["hx-aggrid-cols"] = cols
        self.attributes["hx-aggrid-rows"] = rows
        self.attributes["hx-aggrid"] = "/data"

    @staticmethod
    def create_cols_and_rows(df: pd.DataFrame, escape: bool = True) -> Tuple[list[Any], list[Any]]:
        cols = []
        rows = []

        if df is not None:
            df = df.replace({np.nan: None})
            cols = [{"field": item} for item in df.columns.to_list()]
            rows = df.to_dict("records")
            if escape:
                cols = html.escape(json.dumps(cols))
                rows = html.escape(json.dumps(rows, default=str))

        return cols, rows

    @staticmethod
    def response(df: pd.DataFrame, headers: dict[str, str] = {}) -> JSONResponse:
        _headers = {"HX-Trigger": "aggridUpdate"} | headers
        cols, rows = Aggrid.create_cols_and_rows(df, False)
        return JSONResponse({"cols": cols, "rows": rows}, headers=_headers)



from datetime import datetime
import typing
from fastapi import Response


class LoginResponse(Response):
    def __init__(
        self,
        url: str = "/",
        headers: typing.Optional[typing.Mapping[str, str]] = None,
    ) -> None:
        super().__init__("", 200, headers)
        self.headers["Hx-Redirect"] = url
        self.body = b""

    def set_token(
        self,
        session_id: str,
        session_token: str,
        expires: typing.Union[datetime, str, int],
    ):
        self.set_cookie(
            session_id,
            session_token,
            expires=expires,
            secure=True,
            samesite="strict",
        )



from starlette.middleware.base import BaseHTTPMiddleware


class TtlMiddelware(BaseHTTPMiddleware):
    def __init__(self, app, cache_age: int):
        super().__init__(app)
        self.cahce_age = cache_age

    async def dispatch(self, request, call_next):
        response = await call_next(request)
        if "static/" in str(request.url):
            response.headers["Cache-Control"] = f"max-age={self.cahce_age}"
        return response



import inspect
from mimetypes import guess_type
import os
from pathlib import Path
from typing import Callable, Optional, Union
from fastapi import FastAPI, Response
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.requests import Request
from uiwiz.request_middelware import RequestMiddleware
from uiwiz.element import Element, Frame
from fastapi.middleware.gzip import GZipMiddleware
import functools
import logging
from uiwiz.page_route import PageRouter
from uiwiz.static_middelware import TtlMiddelware
from html import escape

logger = logging.getLogger("uiwiz")
logger.addHandler(logging.NullHandler())


class UiwizApp(FastAPI):
    def __init__(
        self,
        toast_delay: int = 2500,
        error_classes: str = "alert alert-error",
        cache_age: int = 14400,
        theme: Optional[str] = None,
        auth_header: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.toast_delay = toast_delay
        self.error_classes = error_classes
        if theme:
            self.theme = f"data-theme={theme}"
        else:
            self.theme = theme
        self.auth_header = auth_header
        self.templates = Jinja2Templates(Path(__file__).parent / "templates")
        self.add_static_files("/static", Path(__file__).parent / "static")

        self.add_middleware(RequestMiddleware)
        self.add_middleware(GZipMiddleware)
        self.add_middleware(TtlMiddelware, cache_age=cache_age)
        self.extensions: dict[str, Path] = {}
        self.app_paths: dict[str, Path] = {}

    def render(
        self,
        request: Request,
        title: str,
        status_code: int = 200,
    ):
        frame = Frame.get_stack()
        html = frame.render()
        libs = frame.render_libs()
        ext = frame.render_ext()
        theme = self.theme
        if cookie_theme := request.cookies.get("data-theme"):
            theme = f"data-theme={escape(cookie_theme)}"
        return self.return_funtion_response(
            self.templates.TemplateResponse(
                name="default.html",
                context={
                    "request": request,
                    "root_element": [html],
                    "title": title,
                    "theme": theme,
                    "libs": libs,
                    "ext": ext,
                    "toast_delay": self.toast_delay,
                    "error_classes": self.error_classes,
                    "auth_header_name": self.auth_header,
                },
                status_code=status_code,
                headers={"Cache-Control": "no-store", "X-uiwiz-Content": "page"},
            )
        )

    def render_api(self, status_code: int = 200):
        return self.return_funtion_response(
            HTMLResponse(
                Frame.get_stack().render(),
                status_code,
                {"Cache-Control": "no-store", "X-uiwiz-Content": "partial-ui"},
            )
        )

    def route_exists(self, path: str) -> None:
        return path in list(self.app_paths.values())

    def remove_route(self, path: str) -> None:
        """Remove routes with the given path."""
        self.routes[:] = [r for r in self.routes if getattr(r, "path", None) != path]

    def add_static_files(self, url_path: str, local_directory: Union[str, Path]) -> None:
        self.mount(url_path, StaticFiles(directory=str(local_directory)))

    def register_extension(self, path: Path, prefix: str):
        _, filename = os.path.split(path)
        if filename in self.extensions:
            return
        self.extensions[filename] = path

        def get_extension(filename):
            if filename not in self.extensions:
                return Response(status_code=404)

            with open(self.extensions[filename]) as f:
                content = f.read()

            content_type, _ = guess_type(filename)
            return Response(content, media_type=content_type)

        self.get(prefix + "{filename}")(get_extension)

    def post(self, path: str, *args, **kwargs):
        s = super()

        def decorator(func: Callable, *args, **kwargs) -> Callable:
            return s.post(path, *args, **kwargs)(func)

        return decorator

    def page(
        self,
        path: str,
        *args,
        title: Optional[str] = "uiwiz",
        favicon: Optional[str] = None,
    ) -> Callable:
        def decorator(func: Callable, *args, **kwargs) -> Callable:
            self.remove_route(path)
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

            @functools.wraps(func)
            async def decorated(*dec_args, **dec_kwargs) -> Response:
                # Create frame before function is called
                Frame.get_stack()
                Element().classes("flex flex-col min-h-screen h-full")
                request = dec_kwargs["request"]
                # NOTE cleaning up the keyword args so the signature is consistent with "func" again
                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                result = func(*dec_args, **dec_kwargs)
                if inspect.isawaitable(result):
                    result = await result
                if isinstance(result, Response):
                    return self.return_funtion_response(result)

                return self.render(request, title)

            params = [p for p in inspect.signature(func).parameters.values()]
            if "request" not in {p.name for p in params}:
                request = inspect.Parameter(
                    "request",
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=Request,
                )
                params.insert(0, request)
            decorated.__signature__ = inspect.Signature(params)

            if not self.route_exists(path):
                self.app_paths[decorated] = path

            return self.get(path)(decorated)

        return decorator

    def ui(self, path: str) -> Callable:
        def decorator(func: Callable) -> Callable:
            self.remove_route(path)
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

            @functools.wraps(func)
            async def decorated(*dec_args, **dec_kwargs) -> Response:
                # Create frame before function is called
                Frame.get_stack()
                # NOTE cleaning up the keyword args so the signature is consistent with "func" again
                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                result = func(*dec_args, **dec_kwargs)
                if inspect.isawaitable(result):
                    result = await result
                if isinstance(result, Response):  # NOTE if setup returns a response, we don't need to render the page
                    return self.return_funtion_response(result)

                return self.render_api()

            request = inspect.Parameter("request", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=Request)
            params = [p for p in inspect.signature(func).parameters.values()]
            for p in params:
                if p.annotation == inspect.Signature.empty:
                    p._annotation = Request
            if "request" not in {p.name for p in params}:
                params.insert(0, request)
            decorated.__signature__ = inspect.Signature(params)

            if not self.route_exists(path):
                self.app_paths[decorated] = path

            return self.post(path)(decorated)

        return decorator

    def include_page_router(self, page_router: PageRouter):
        for key, value in page_router.paths.items():
            if not self.route_exists(key):
                self.app_paths[value.get("func")] = key
            type = value.get("type")
            if type == "page":
                self.page(key)(value.get("func"))
            if type == "ui":
                self.ui(key)(value.get("func"))

    def return_funtion_response(self, response: Response) -> Response:
        Frame.get_stack().del_stack()
        return response



from typing import Callable, Optional


class PageRouter:
    def __init__(self) -> None:
        self.paths = {}

    def page(
        self,
        path: str,
        *args,
        title: Optional[str] = "uiwiz",
        favicon: Optional[str] = None,
    ) -> Callable:
        def setup(func, *args, **kwargs):
            self.paths[path] = {"func": func, "type": "page"}
            return func

        return setup

    def ui(self, path: str) -> Callable:
        def setup(func, *args, **kwargs):
            self.paths[path] = {"func": func, "type": "ui"}
            return func

        return setup



from typing import Callable, Optional, TypedDict


class Event(TypedDict):
    func: Callable
    trigger: str
    target: Optional[str]
    swap: Optional[str]
    include: Optional[str]
    vals: Optional[str]
    params: Optional[dict]



from datetime import datetime
from typing import Optional
from uiwiz.element import Element


class Datepicker(Element):
    def __init__(self, name: str, default_date: Optional[datetime] = None) -> None:
        super().__init__("input")
        self.attributes["name"] = name
        self.attributes["type"] = "date"
        if default_date:
            self.attributes["value"] = default_date.strftime("%Y-%m-%d")



from fastapi import Request
from uiwiz.element import Element


class Bindable(Element):
    def bind_text_from(self, element: Element, trigger: str = "input delay:20ms", swap: str = "outerHTML"):
        """
        Remove attribute as it is not used but breaks bind for reasons
        """
        assert element.attributes.get("name") is not None

        async def bind_value(request: Request):
            data = await request.json()
            self.set_frame_and_root()

            input_data = data[element.name]

            self.content = str(input_data)

        bind_value.__hash__ = lambda: hash(element.id)

        element.event = {"target": f"#{self.id}", "func": bind_value, "trigger": trigger, "swap": swap}

        return self



from typing import Optional
from uiwiz.element import Element


class Tabs(Element):
    root_class: str = "tabs "
    _classes: str = "tabs-bordered"

    def __init__(self) -> None:
        super().__init__()
        self.classes(Tabs.root_class + Tabs._classes)
        self.attributes["role"] = "tablist"

    def __exit__(self, *_):
        super().__exit__(*_)
        has_active = False
        first_child_tab = None
        for child in self.children:
            if not isinstance(child, Tab):
                continue

            if first_child_tab is None:
                first_child_tab = child

            if "checked" in child.attributes:
                has_active = True
        if not has_active:
            first_child_tab.active()


class Tab(Element):
    _classes: str = "tab"

    def __init__(self, title: str, active: Optional[bool] = None) -> None:
        self.selector = Element("input")
        self.selector.attributes["name"] = self.selector.parent_element.id
        self.selector.attributes["type"] = "radio"
        self.selector.attributes["aria-label"] = title
        self.selector.classes(self._classes)
        if active:
            self.active()

        super().__init__()
        self.classes("tab-content bg-base-100 border-base-300 p-4")
        self.attributes["role"] = "tabpanel"

    def active(self) -> "Tab":
        self.selector.attributes["checked"] = ""
        return self



from uiwiz.elements.nav import Nav as nav
from uiwiz.elements.button import Button as button
from uiwiz.elements.footer import Footer as footer
from uiwiz.elements.label import Label as label
from uiwiz.elements.toast import Toast as toast
from uiwiz.elements.markdown import Markdown as markdown
from uiwiz.elements.input import Input as input
from uiwiz.elements.textarea import TextArea as textarea
from uiwiz.elements.table import Table as table
from uiwiz.elements.aggrid import Aggrid as aggrid
from uiwiz.elements.drawer import Drawer as drawer
from uiwiz.elements.link import Link as link
from uiwiz.elements.upload import Upload as upload
from uiwiz.elements.checkbox import Checkbox as checkbox
from uiwiz.elements.dropdown import Dropdown as dropdown
from uiwiz.elements.datepicker import Datepicker as datepicker
from uiwiz.elements.toggle import Toggle as toggle
from uiwiz.elements.code import Code as code
from uiwiz.elements.avatar import Avatar as avatar
from uiwiz.elements.tabs import Tabs as tabs
from uiwiz.elements.tabs import Tab as tab
from uiwiz.elements.divider import Divider as divider
from uiwiz.elements.form import Form as form
from uiwiz.elements.radio import Radio as radio
from uiwiz.elements.range import Range as range
from uiwiz.elements.row import Row as row
from uiwiz.elements.col import Col as col
from uiwiz.elements.spinner import Spinner as spinner
from uiwiz.elements.html import Html as html
from uiwiz.elements.theme_selector import ThemeSelector as themeSelector
from uiwiz.element import Element as element



from uiwiz.elements.html import Html
import markdown2


class Code(Html):
    def __init__(self, code: str) -> None:
        super().__init__("div")
        color = "bg-neutral-content"
        _class = "mockup-window " + color

        self.classes(_class)

        self.content = markdown2.markdown(code, extras=["fenced-code-blocks"])
        self.content = self.content.replace("<pre>", f'<pre class="{color}"')



from contextvars import ContextVar
from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware

REQUEST_CTX_KEY = "_request"

_request_ctx_var: ContextVar[dict[str, Request]] = ContextVar(REQUEST_CTX_KEY, default=None)


def get_request() -> Request:
    return _request_ctx_var.get()


class RequestMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        _request = _request_ctx_var.set(request)
        response = await call_next(request)
        _request_ctx_var.reset(_request)

        return response



from uiwiz.element import Element
from uiwiz import ui
from uiwiz.svg.svg_handler import get_svg, _type


class Toast(Element):
    root_class: str = "alert w-full z-50 "

    def __init__(self, message: str = "", svg: _type = None) -> None:
        super().__init__(tag="div", oob=True)
        self.attributes["id"] = "toast"
        self.attributes["hx-swap-oob"] = "afterbegin"
        self._svg = svg
        self.message = message
        self.context_manager_used = False
        self.inner_element = None
        self.inner_class = Toast.root_class

    def before_render(self):
        with self:
            with self.inner_element.classes(self.inner_class):
                if self._svg:
                    svg = ui.html(get_svg(self._svg))
                    self.inner_element.children.remove(svg)
                    self.inner_element.children.insert(0, svg)
                Element("span", content=self.message).inline = True

    def __enter__(self):
        super().__enter__()
        if self.inner_element is None:
            self.inner_element = Element().classes(self.inner_class)
        self.inner_element.__enter__()
        return self

    def svg(self, svg: _type) -> "Toast":
        self._svg = svg
        return self

    def classes(self, input: str) -> "Toast":
        self.inner_class = self.inner_class + input
        return self

    def info(self) -> "Toast":
        self.classes("alert")
        self.svg("info")
        return self

    def warning(self) -> "Toast":
        self.classes("alert-warning")
        self.svg("warning")
        return self

    def success(self) -> "Toast":
        self.classes("alert-success")
        self.svg("success")
        return self

    def error(self) -> "Toast":
        self.classes("alert-error")
        self.svg("error")
        return self



import numpy as np
from uiwiz.element import Element
import pandas as pd


class Table(Element):
    _classes_container: str = "w-full overflow-x-auto uiwiz-container-border-radius"
    _classes_table: str = "table-zebra table-auto bg-base-300 overflow-scroll w-full whitespace-nowrap uiwiz-td-padding"

    def __init__(self, df: pd.DataFrame) -> None:
        super().__init__()
        self.classes(Table._classes_container)
        df = df.replace({np.nan: "None"})

        with self:
            with Element("table").classes(Table._classes_table):
                # columns
                with Element("thead"):
                    with Element("tr"):
                        for col in df.columns:
                            Element("th", content=col)
                # rows
                with Element("tbody"):
                    for _, row in df.iterrows():
                        with Element("tr"):
                            for _, val in row.items():
                                Element("td", content=val)
